/*
 * Copyright (C) 2009 SENJU, Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE
#define _REENTRANT

#include <pthread.h>

#include <sys/epoll.h>
#include <sys/time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define __LAIO_LIB__ /* users never use this */
#include "laio.h"

struct la_counter {
	pthread_mutex_t mtx;
	int cur, lim;
};

static struct {
	aio_context_t ctx;
	int epfd;
	struct la_counter nthread, nio;

	struct {
		pthread_mutex_t mtx;
		pthread_cond_t cond;
	} suspend;
} laio = {
	/* values borrowed from glibc.git/sysdeps/pthread/aio_misc.c */
	.nthread	= {
		.mtx	= PTHREAD_MUTEX_INITIALIZER,
		.lim	= 20
	},

	.nio		= {
		.mtx	= PTHREAD_MUTEX_INITIALIZER,
		.lim	= 64
	},

	.suspend = {
		.mtx	= PTHREAD_MUTEX_INITIALIZER,
		.cond	= PTHREAD_COND_INITIALIZER
	}
};

#ifndef offsetof
#define offsetof(TYPE, MEMBER)  __builtin_offsetof (TYPE, MEMBER)
#endif

/* borrowed from linux/kernel.h */
/**
 * container_of - cast a member of a structure out to the containing structure
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})

/* ---------------------------------------------------------------------- */

#define DPri(fmt, args...) \
	fprintf(stderr, "%s:%d:[%d]:" fmt, \
		__func__, __LINE__, getpid(), ##args)

/* ---------------------------------------------------------------------- */

static int la_cnt_read(struct la_counter *cnt)
{
	int v;

	pthread_mutex_lock(&cnt->mtx);
	v = cnt->cur;
	pthread_mutex_unlock(&cnt->mtx);

	return v;
}

static void la_cnt_inc(struct la_counter *cnt)
{
	pthread_mutex_lock(&cnt->mtx);
	cnt->cur++;
	pthread_mutex_unlock(&cnt->mtx);
}

static void la_cnt_dec(struct la_counter *cnt)
{
	pthread_mutex_lock(&cnt->mtx);
	cnt->cur--;
	pthread_mutex_unlock(&cnt->mtx);
}

/* ---------------------------------------------------------------------- */

static int la_create_thread(void *(*func)(void *), pthread_attr_t *attr,
			      void *arg)
{
	int err;
	pthread_attr_t _attr;
	pthread_t th;

	err = 0;
	if (!attr) {
		attr = &_attr;
		err = pthread_attr_init(attr);
		if (!err)
			err = pthread_attr_setdetachstate
				(attr, PTHREAD_CREATE_DETACHED);
	}
	if (!err) {
		if (la_cnt_read(&laio.nthread) < laio.nthread.lim) {
			err = pthread_create(&th, attr, func, arg);
			if (!err)
				la_cnt_inc(&laio.nthread);
		} else {
			err = -1;
			errno = EAGAIN;
		}
	}
	return err;
}

/* ---------------------------------------------------------------------- */

struct la_th_data {
	void (*func)(sigval_t v);
	sigval_t val;
};

static void *la_notify_thread(void *_arg)
{
	struct la_th_data arg, *p = _arg;

	arg = *p;
	free(p);
	arg.func(arg.val);
	la_cnt_dec(&laio.nthread);
	return NULL;
}

static int la_notify(pid_t pid, struct sigevent *se)
{
	int err;

	//DPri("se %p{v, no %d, notify %d}\n", se, se->sigev_signo, se->sigev_notify);
	err = 0;
	if (se->sigev_notify == SIGEV_SIGNAL)
		err = sigqueue(pid, se->sigev_signo, se->sigev_value);
	else if (se->sigev_notify == SIGEV_THREAD) {
		struct la_th_data *arg;

		err = -1;
		arg = malloc(sizeof(*arg));
		if (!arg)
			goto out;

		arg->func = se->sigev_notify_function;
		arg->val = se->sigev_value;
		err = la_create_thread(la_notify_thread,
				       se->sigev_notify_attributes, arg);
	}

 out:

	//DPri("err %d\n", err);
	return err;
}

static void *la_getevents(void *arg)
{
	int err, i, n;
	struct io_event ev[laio.nio.lim], *e;
	struct la_cb *cb;
	pid_t pid;

	pid = getpid();
	while (1) {
		//pthread_mutex_lock(&laio.suspend.mtx);
		n = io_getevents(laio.ctx, /*min_nr*/1, /*nr*/laio.nio.lim, ev,
				 /*struct timespec*/NULL);
		//DPri("n %d\n", n);
		if (n < 0)
			abort();

		for (e = ev, i = 0; i < n; i++, e++) {
			la_cnt_dec(&laio.nio);
			//DPri("%llu, 0x%llx, %lld, %lld\n", e->data, e->obj, e->res, e->res2);
			cb = container_of((void *)(unsigned long)e->obj,
					  struct la_cb, la_iocb);
			pthread_mutex_lock(&laio.suspend.mtx);
			cb->la_aiocb.__return_value = e->res;
			if (e->res < 0) {
				cb->la_aiocb.__return_value = -1;
				cb->la_aiocb.__error_code = -e->res;
			}
			//DPri("broadcast\n");
			pthread_cond_broadcast(&laio.suspend.cond);
			pthread_mutex_unlock(&laio.suspend.mtx);

			err = la_notify(pid, &cb->la_aiocb.aio_sigevent);
			//DPri("err %d\n", err);
			if (err < 0)
				abort();
		}
		//pthread_mutex_unlock(&laio.suspend.mtx);
	}

	return NULL;
}

static void *la_epoll(void *arg)
{
	int err, i, n;
	struct epoll_event ev[laio.nio.lim];
	struct la_cb *cb;
	struct iocb *iocb;

	while (1) {
		err = epoll_wait(laio.epfd, ev, laio.nio.lim, -1);
		if (err)
			abort();

		n = err;
		for (i = 0; i < n; i++) {
			cb = ev[i].data.ptr;
			iocb = &cb->la_iocb;
			err = io_submit(laio.ctx, /*nr*/1, &iocb);
			if (err)
				fprintf(stderr, "%s: internal error, "
					"failed io_submit\n", __func__);
		}
	}

	return NULL;
}

/* ---------------------------------------------------------------------- */

static int la_do_init(int nthread, int nio)
{
	int err, eno;
	static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;

	err = 0;
	pthread_mutex_lock(&mtx);
	if (laio.ctx)
		goto out;

	//DPri("here\n");
	err = -1;
	laio.nthread.lim = nthread;
	if (laio.nthread.lim < 0)
		laio.nthread.lim = 0;
	laio.nio.lim = nio;
	if (laio.nio.lim < 1)
		laio.nio.lim = 1;

	err = io_setup(laio.nio.lim, &laio.ctx);
	if (err)
		goto out;

	if (!laio.ctx) {
		fprintf(stderr, "%s: internal error, unexpected zero\n",
			__func__);
		goto out;
	}

	err = epoll_create(1);
	if (err < 0)
		goto out;
	laio.epfd = err;
	err = la_create_thread(la_epoll, NULL, NULL);
	if (err)
		goto out;

	if (laio.nthread.lim)
		err = la_create_thread(la_getevents, NULL, NULL);

 out:
	eno = errno;
	pthread_mutex_unlock(&mtx);
	errno = eno;
	return err;
}

#ifdef __USE_GNU
void laio_init(const struct aioinit *init)
{
	int err;

	if (laio.ctx)
		return;

	err = la_do_init(init->aio_threads, init->aio_num);
	if (err)
		abort();
}
#endif

/* ---------------------------------------------------------------------- */

static int la_rw(struct la_cb *cb, __u16 cmd)
{
	int err, do_epoll;
	struct iocb *iocb;
	unsigned long ul;
	struct epoll_event ev;

	err = -1;
	if (!laio.ctx) {
		err = la_do_init(laio.nthread.lim, laio.nio.lim);
		if (err)
			goto out;
		err = -1;
	}

	ul = (unsigned long)cb->la_aiocb.aio_buf;
	iocb = &cb->la_iocb;
	memset(iocb, 0, sizeof(*iocb));
	iocb->aio_lio_opcode = cmd;
	iocb->aio_fildes = cb->la_aiocb.aio_fildes;
	iocb->aio_buf = (__u64)ul;
	iocb->aio_nbytes = cb->la_aiocb.aio_nbytes;
	iocb->aio_offset = cb->la_aiocb.aio_offset;
	laio_eventfd(cb);
	cb->la_aiocb.__return_value = -EINPROGRESS;

	do_epoll = 0;
	if (cmd == IOCB_CMD_PREAD) {
		ev.events = EPOLLIN | EPOLLHUP | EPOLLPRI | EPOLLONESHOT;
		do_epoll = 1;
	} else if (cmd == IOCB_CMD_PWRITE) {
		ev.events = EPOLLOUT | EPOLLHUP | EPOLLONESHOT;
		do_epoll = 1;
	}
	if (do_epoll) {
		/*
		 * the success of epoll_ctl() means that the poll
		 * operation of fd is implemented, and we should use it before
		 * io_submit().
		 * otherwise the poll is not implemented, and we should call
		 * io_submit() at once.
		 */
		ev.data.ptr = cb;
		err = epoll_ctl(laio.epfd, EPOLL_CTL_ADD, iocb->aio_fildes,
				&ev);
		if (!err) {
			la_cnt_inc(&laio.nio);
			goto out;
		}
	}

	err = io_submit(laio.ctx, /*nr*/1, &iocb);
	//DPri("err %d\n", err);
	if (err == 1) {
		err = 0;
		la_cnt_inc(&laio.nio);
	} else {
		cb->la_aiocb.__return_value = err;
		cb->la_aiocb.__error_code = errno;
	}

 out:
	return err;
}

int laio_read(struct la_cb *cb)
{
	return la_rw(cb, IOCB_CMD_PREAD);
}

int laio_write(struct la_cb *cb)
{
	return la_rw(cb, IOCB_CMD_PWRITE);
}

int laio_fsync(int op, struct la_cb *cb)
{
	int flags;

	errno = EBADF;
	flags = fcntl(cb->la_aiocb.aio_fildes, F_GETFL);
	if (flags == -1 || !(flags & (O_RDWR | O_WRONLY)))
		return -1;

	//DPri("buf %p, nb %u\n", cb->la_aiocb.aio_buf, cb->la_aiocb.aio_nbytes);
	//cb->la_aiocb.aio_nbytes = 0;
	if (op == O_SYNC)
		return la_rw(cb, IOCB_CMD_FSYNC);
	else if (op == O_DSYNC)
		return la_rw(cb, IOCB_CMD_FDSYNC);
	else {
		errno = EINVAL;
		return -1;
	}
}

ssize_t laio_return(struct la_cb *cb)
{
	ssize_t err;

	err = cb->la_aiocb.__return_value;
#if 0
	if (err < 0)
		err *= -1;
#endif
	return err;
}

int laio_error(const struct la_cb *cb)
{
	int err;

	err = cb->la_aiocb.__error_code;
	//err = cb->la_aiocb.__return_value;
#if 0
	if (err >= 0)
		err = 0;
	else
		err *= -1;
#endif
	return err;
}

static int la_suspend_test(const struct la_cb * const l[], int n)
{
	int err;
	int i;

	err = 0;
	for (i = 0; i < n; i++)
		if (l[i]) {
			if (l[i]->la_aiocb.__return_value == -EINPROGRESS)
				err = -1;
			break;
		}
	return err;
}

int laio_suspend(const struct la_cb * const cbpl[], int n,
		 const struct timespec *timeout)
{
	int err, to;

	errno = EINVAL;
	err = -1;
	to = 0;
	if (!laio.ctx || n > laio.nio.lim)
		goto out;

	//DPri("here\n");
	err = 0;
	pthread_mutex_lock(&laio.suspend.mtx);
	while (to != ETIMEDOUT && la_cnt_read(&laio.nio) > 0) {
		err = la_suspend_test(cbpl, n);
		//DPri("err %d\n", err);
		if (!err)
			break;

		if (timeout == NULL) {
			err = pthread_cond_wait(&laio.suspend.cond, &laio.suspend.mtx);
			if (!err)
				to = ETIMEDOUT;
			else
				break;
		} else {
			/* convert the relative timeout into absolute */
			struct timeval now;
			struct timespec abs;

			gettimeofday(&now, NULL);
			abs.tv_nsec = timeout->tv_nsec + now.tv_usec * 1000;
			abs.tv_sec = timeout->tv_sec + now.tv_sec;
			if (abs.tv_nsec >= 1000000000) {
				abs.tv_nsec -= 1000000000;
				abs.tv_sec += 1;
			}
			to = pthread_cond_timedwait
				(&laio.suspend.cond, &laio.suspend.mtx, &abs);
		}
	}
	pthread_mutex_unlock(&laio.suspend.mtx);

 out:
	//DPri("here\n");
	if (to == ETIMEDOUT)
		errno = EAGAIN;
	//DPri("err %d\n", err);
	return err;
}

static int la_cancel_one(struct la_cb *cb)
{
	int err;
	struct io_event ev;

	err = io_cancel(laio.ctx, &cb->la_iocb, &ev);
	if (!err) {
		if (cb->la_aiocb.__return_value == -EINPROGRESS) {
			err = AIO_CANCELED;
			cb->la_aiocb.__return_value = -1;
			cb->la_aiocb.__error_code = ECANCELED;
			la_cnt_dec(&laio.nio);
		} else
			err = AIO_ALLDONE;
	}

	return err;
}

/*
 * Since io_cancel(2) doesn't distinguish the case of already completed,
 * this function doesn't return AIO_ALLDONE either.
 */
int laio_cancel(int fd, struct la_cb *cb)
{
	int err, eno/* , idx */;

	eno = EINVAL;
	err = -1;
	if (cb) {
		if (fd < 0 || fd != cb->la_aiocb.aio_fildes) {
			eno = EBADF;
			goto out;
		}
		err = la_cancel_one(cb);
	} else {
		if (fd < 0) {
			eno = EBADF;
			goto out;
		}
#if 0
		/* unsupported */
		err = AIO_CANCELED;
		idx = 0;
		while (idx < laio.nio) {
			cb = la_cb_get_fd(fd, &idx);
			if (cb) {
				if (la_cancel_one(cb))
					err = AIO_NOTCANCELED;
			} else
				break;
		}
#endif
	}

 out:
	errno = eno;
	return err;
}

static int lio_wait(struct la_cb *cb)
{
	int err;
	const struct la_cb * const l[] = {cb};

	pthread_mutex_lock(&laio.suspend.mtx);
	while (1) {
		err = la_suspend_test(l, 1);
		if (!err)
			break;
		pthread_cond_wait(&laio.suspend.cond, &laio.suspend.mtx);
	}
	pthread_mutex_unlock(&laio.suspend.mtx);

	return err;
}

int llio_listio(int mode, struct la_cb * const cbpl[], int nent,
		struct sigevent *sig)
{
	int err, i;
	struct la_cb *cb;
	struct sigevent s;

	if (mode == LIO_WAIT)
		sig = NULL;
	else if (sig)

	err = -1;
	for (i = 0; i < nent; i++) {
		cb = cbpl[i];
		if (sig) {
			s = cb->la_aiocb.aio_sigevent;
			cb->la_aiocb.aio_sigevent = *sig;
		}
		switch (cb->la_aiocb.aio_lio_opcode) {
		case LIO_READ:
			err = la_rw(cb, IOCB_CMD_PREAD);
			break;
		case LIO_WRITE:
			err = la_rw(cb, IOCB_CMD_PWRITE);
			/* FALLTHROUGH */
		case LIO_NOP:
			break;
		default:
			err = -1;
			errno = EINVAL;
			goto out;
		}
		if (sig)
			cb->la_aiocb.aio_sigevent = s;
		if (err)
			goto out;
	}

	if (mode == LIO_WAIT)
		for (i = 0; i < nent; i++)
			if (cbpl[i])
				err = lio_wait(cbpl[i]);

 out:
	return err;
}

/* ---------------------------------------------------------------------- */

#ifdef __LAIO_LIB64__ /* users never use this */
int laio_read64(struct la_cb64 *cb)
{
	return laio_read(cb);
}

int laio_write64(struct la_cb64 *cb)
{
	return laio_write(cb);
}

int laio_fsync64(int op, struct la_cb64 *cb)
{
	return laio_fsync(op, cb);
}

ssize_t laio_return64(struct la_cb64 *cb)
{
	return laio_return(cb);
}

int laio_error64(const struct la_cb64 *cb)
{
	return laio_error(cb);
}

int laio_suspend64(const struct la_cb64 * const cbpl[], int n,
		   const struct timespec *timeout)
{
	return laio_suspend(cbpl, n, timeout);
}

int laio_cancel64(int fd, struct la_cb64 *cb)
{
	return laio_cancel(fd, cb);
}

int llio_listio64(int mode, struct la_cb64 * const cbpl[], int nent,
		  struct sigevent *sig)
{
	return llio_listio(mode, cbpl, nent, sig);
}
#endif
